#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<signal.h>
#include	<unistd.h>
#include	<errno.h>
#include	<sys/ipc.h>
#include	<sys/msg.h>
#include	<pthread.h>

#define	PARAM_PATH	"./logger.ini"
#define	MAX_SIZE	(8192)

typedef struct	{
	char	*name;
	int	qno;
	char	*path;
	char	rotateType;
	int	sizeLimit;
}PARAM;

PARAM	*Param=NULL;
int	ParamNo=0;

typedef struct	{
	pthread_t	tid;
	char	*fname;
	int	qid;
	FILE	*fp;
	int	reopen;
	int	year,mon,mday,hour;
}THREAD_DATA;

THREAD_DATA	*TD=NULL;
int	TDNo=0;

pthread_t	ParentTid;
int	EndFlag=0;

int ReadParam(char *fname)
{
FILE	*fp;
PARAM	p;
char	buf[2048],*ptr,*saveptr;

	if((fp=fopen(fname,"r"))==NULL){
		perror("fopen");
		return(-1);
	}

	while(1){
		if(fgets(buf,sizeof(buf),fp)==NULL){
			break;
		}
		if(buf[0]=='#'){
			continue;
		}
		if((ptr=strtok_r(buf,",",&saveptr))==NULL){
			continue;
		}
		p.name=ptr;
		if((ptr=strtok_r(NULL,",",&saveptr))==NULL){
			continue;
		}
		p.qno=atoi(ptr);
		if((ptr=strtok_r(NULL,",",&saveptr))==NULL){
			continue;
		}
		p.path=ptr;
		if((ptr=strtok_r(NULL,",",&saveptr))==NULL){
			continue;
		}
		if(*ptr=='N'||*ptr=='S'||*ptr=='H'){
			p.rotateType=*ptr;
		}
		else{
			continue;
		}
		if((ptr=strtok_r(NULL,"\r\n",&saveptr))==NULL){
			continue;
		}
		p.sizeLimit=atoi(ptr);
		if(ParamNo==0){
			Param=(PARAM *)malloc(sizeof(PARAM));
		}
		else{
			Param=(PARAM *)realloc(Param,(ParamNo+1)*sizeof(PARAM));
		}
		Param[ParamNo].name=strdup(p.name);
		Param[ParamNo].qno=p.qno;
		Param[ParamNo].path=strdup(p.path);
		Param[ParamNo].rotateType=p.rotateType;
		Param[ParamNo].sizeLimit=p.sizeLimit;
		printf("[%d]:%s,%d,%s,%c,%d\n",
			ParamNo,
			Param[ParamNo].name,
			Param[ParamNo].qno,
			Param[ParamNo].path,
			Param[ParamNo].rotateType,
			Param[ParamNo].sizeLimit);
		ParamNo++;
	}

	fclose(fp);

	return(0);
}

int checkRotatePre(int no)
{
THREAD_DATA	*td=&TD[no];
char	*bakname;

	if(Param[no].rotateType=='H'){
		time_t	t;
		struct tm	tm;
		t=time(NULL);
		localtime_r(&t,&tm);
		if(td->hour==-1){
			td->year=tm.tm_year+1900;
			td->mon=tm.tm_mon+1;
			td->mday=tm.tm_mday;
			td->hour=tm.tm_hour;
			return(0);
		}
		else if(td->hour!=tm.tm_hour){
			if(td->fp!=NULL){
				fclose(td->fp);
			}
			bakname=(char *)calloc(strlen(td->fname)+1+4+2+2+2+1,sizeof(char));
			sprintf(bakname,"%s.%04d%02d%02d%02d",td->fname,td->year,td->mon,td->mday,td->hour);
			rename(td->fname,bakname);
			free(bakname);
			td->year=tm.tm_year+1900;
			td->mon=tm.tm_mon+1;
			td->mday=tm.tm_mday;
			td->hour=tm.tm_hour;
			td->fp=fopen(td->fname,"a");
			return(1);
		}
		else{
			return(0);
		}
	}
	else{
		return(0);
	}
}

int checkRotatePost(int no)
{
THREAD_DATA	*td=&TD[no];
char	*bakname;

	if(Param[no].rotateType=='N'){
		return(0);
	}
	else if(Param[no].rotateType=='S'){
		long	size;
		size=ftell(td->fp);
		if(size>Param[no].sizeLimit*1024*1024){
			if(td->fp!=NULL){
				fclose(td->fp);
			}
			bakname=(char *)calloc(strlen(td->fname)+1+3+1,sizeof(char));
			sprintf(bakname,"%s.bak",td->fname);
			rename(td->fname,bakname);
			free(bakname);
			td->fp=fopen(td->fname,"a");
			return(1);
		}
		else{
			return(0);
		}
	}
	else{
		return(0);
	}
}

void *loggerThread(void *arg)
{
int	no=(int)arg;
THREAD_DATA	*td=&TD[no];
struct msgbuf{
	long	mtype;
	char	mtext[MAX_SIZE];
}buf;
ssize_t	size;

	while(EndFlag==0){
		if(td->reopen==1){
			if(td->fp!=NULL){
				fclose(td->fp);
			}
			td->fp=fopen(td->fname,"a");
			td->reopen=0;
		}
		if((size=msgrcv(td->qid,&buf,sizeof(buf.mtext),0L,MSG_NOERROR))==-1){
			perror("msgrcv");
		}
		else{
			checkRotatePre(no);
			if(td->fp!=NULL){
				fwrite(buf.mtext,sizeof(char),size,td->fp);
				fputc('\n',td->fp);
			}
			checkRotatePost(no);
		}
	}

	fclose(td->fp);

	return(0);
}

void ending(int sig)
{
	if(pthread_self()==ParentTid){
		EndFlag=1;
	}
}

void reopen(int sig)
{
int	i;

	if(pthread_self()==ParentTid){
		for(i=0;i<ParamNo;i++){
			TD[i].reopen=1;
			pthread_kill(TD[i].tid,SIGUSR1);
		}
	}
}

int main(int argc,char *argv[])
{
int	i,ret;

	if(argc==1){
		if(ReadParam(PARAM_PATH)==-1){
			return(-1);
		}
	}
	else{
		if(ReadParam(argv[1])==-1){
			return(-1);
		}
	}

	TD=(THREAD_DATA *)calloc(ParamNo,sizeof(THREAD_DATA));

	ParentTid=pthread_self();

	signal(SIGINT,ending);
	signal(SIGTERM,ending);
	signal(SIGQUIT,ending);
	signal(SIGUSR1,reopen);
	signal(SIGPIPE,SIG_IGN);

	for(i=0;i<ParamNo;i++){
		if((TD[i].qid=msgget(Param[i].qno,0666|IPC_CREAT))==-1){
			perror("msgget");
			return(-1);
		}
		TD[i].fname=(char *)calloc(strlen(Param[i].path)+1+strlen(Param[i].name)+1+3+1,sizeof(char));
		sprintf(TD[i].fname,"%s/%s.log",Param[i].path,Param[i].name);
		if((TD[i].fp=fopen(TD[i].fname,"a"))==NULL){
			perror(TD[i].fname);
			return(-1);
		}
		TD[i].hour=-1;
		if((ret=pthread_create(&TD[i].tid,NULL,loggerThread,i))!=0){
			errno=ret;
			perror("pthread_create");
			return(-1);
		}
	}

	while(EndFlag==0){
		for(i=0;i<ParamNo;i++){
			fflush(TD[i].fp);
		}
		sleep(1);
	}

	for(i=0;i<ParamNo;i++){
		pthread_kill(TD[i].tid,SIGINT);
	}

	for(i=0;i<ParamNo;i++){
		void	*retval;
		pthread_join(TD[i].tid,&retval);
	}

	return(0);
}
